@@ -79,11 +79,13 @@
 #define PyBytes_GET_SIZE PyString_GET_SIZE
 #define PyBytes_Size PyString_Size
 #define PyLong_AS_LONG PyInt_AS_LONG
-#define PyNumber_Long PyNumber_Int
 #define _PyBytes_Resize _PyString_Resize
 #define PyObject_Bytes PyObject_Str
 #endif
 
+/* Cache the values of {cP,p}ickle.{load,dump}s */
+static PyObject *_PylibMC_pickle_loads = NULL;
+static PyObject *_PylibMC_pickle_dumps = NULL;
 
 /* {{{ Type methods */
 static PylibMC_Client *PylibMC_ClientType_new(PyTypeObject *type,
@@ -102,6 +104,28 @@ static PylibMC_Client *PylibMC_ClientType_new(PyTypeObject *type,
     return self;
 }
 
+/* Helper: detect whether the serialize and deserialize methods were overridden. */
+static int _PylibMC_method_is_overridden(PylibMC_Client *self, const char *method) {
+    /* `self.__class__.serialize is _pylibmc.client.serialize`, in C */
+    PyObject *base_method = NULL, *current_class = NULL, *current_method = NULL;
+
+    base_method = PyObject_GetAttrString((PyObject *) &PylibMC_ClientType, method);
+    current_class = PyObject_GetAttrString((PyObject *) self, "__class__");
+    if (current_class != NULL) {
+        current_method = PyObject_GetAttrString(current_class, method);
+    }
+
+    Py_XDECREF(base_method);
+    Py_XDECREF(current_class);
+    Py_XDECREF(current_method);
+
+    if (base_method && current_class && current_method) {
+        return base_method == current_method;
+    } else {
+        return -1;
+    }
+}
+
 static void PylibMC_ClientType_dealloc(PylibMC_Client *self) {
     if (self->mc != NULL) {
 #if LIBMEMCACHED_WITH_SASL_SUPPORT
@@ -180,6 +204,18 @@ static int PylibMC_Client_init(PylibMC_Client *self, PyObject *args,
         }
     }
 
+    /* Detect whether we should dispatch to user-modified Python
+     * serialization implementations. */
+    int native_serialization, native_deserialization;
+    if ((native_serialization = _PylibMC_method_is_overridden(self, "serialize")) == -1) {
+        goto error;
+    }
+    self->native_serialization = (uint8_t) native_serialization;
+    if ((native_deserialization = _PylibMC_method_is_overridden(self, "deserialize")) == -1) {
+        goto error;
+    }
+    self->native_deserialization = (uint8_t) native_deserialization;
+
     while ((c_srv = PyIter_Next(srvs_it)) != NULL) {
         unsigned char stype;
         char *hostname;
@@ -397,16 +433,14 @@ static int _PylibMC_Inflate(char *value, size_t size,
             }
         /* Fall-through */
         case Z_OK:
-            if((tryrealloc = realloc(out, rvalsz << 1)) == NULL || errno == ENOMEM) {
+            if ((tryrealloc = realloc(out, rvalsz << 1)) == NULL) {
                 *failure_reason = "realloc";
                 rc = Z_MEM_ERROR;
                 goto zerror;
             }
-
             out = tryrealloc;
 
             /* Wind forward */
-
             strm.next_out = (unsigned char*)(out + rvalsz);
             strm.avail_out = rvalsz;
             rvalsz = rvalsz << 1;
@@ -422,15 +456,11 @@ static int _PylibMC_Inflate(char *value, size_t size,
         goto error;
     }
 
-    tryrealloc = realloc(out, strm.total_out);
-
-    if(tryrealloc == NULL || errno == ENOMEM) {
-        /* we failed to *shrink* the value? */
+    if ((tryrealloc = realloc(out, strm.total_out)) == NULL) {
         *failure_reason = "realloc";
         rc = Z_MEM_ERROR;
         goto error;
     }
-
     out = tryrealloc;
 
     *result = out;
@@ -442,7 +472,7 @@ zerror:
     inflateEnd(&strm);
 
 error:
-    if(out != NULL) {
+    if (out != NULL) {
         free(out);
     }
     *result = NULL;
@@ -451,12 +481,19 @@ error:
 #endif
 /* }}} */
 
-/* {{{ str/bytes key functionality (unicode/bytes on Python 2) */
-static PyObject *_PylibMC_map_str_keys(PyObject *keys) {
+/* Helper for multiset / multiget:
+   1. Take the iterable `keys` and build a map of UTF-8 encoded bytestrings
+      to Unicode keys.
+   2. If `key_array` and `nkeys` are not NULL, additionally copy *new*
+      references to everything in the iterable into `key_array`. Store
+      the actual number of items in `nkeys`.
+*/
+static PyObject *_PylibMC_map_str_keys(PyObject *keys, PyObject **key_array, Py_ssize_t *nkeys) {
     PyObject *key_str_map = NULL;
     PyObject *iter = NULL;
     PyObject *key = NULL;
     PyObject *key_bytes = NULL;
+    Py_ssize_t i = 0;
 
     key_str_map = PyDict_New();
     if (key_str_map == NULL)
@@ -473,13 +510,26 @@ static PyObject *_PylibMC_map_str_keys(PyObject *keys) {
             PyDict_SetItem(key_str_map, key_bytes, key);
             Py_DECREF(key_bytes);
         }
-        Py_DECREF(key);
+        /* stash our owned reference to `key` in this array: */
+        if (key_array != NULL && i < *nkeys) {
+            key_array[i++] = key;
+        } else {
+            Py_DECREF(key);
+        }
+    }
+    if (nkeys != NULL) {
+        *nkeys = i;
     }
 
     Py_DECREF(iter);
     return key_str_map;
 
 cleanup:
+    if (key_array != NULL) {
+        for (Py_ssize_t j = 0; j < i; j++) {
+            Py_DECREF(key_array[j]);
+        }
+    }
     Py_XDECREF(key);
     Py_XDECREF(iter);
     Py_XDECREF(key_str_map);
@@ -491,11 +541,9 @@ static void _PylibMC_cleanup_str_key_mapping(PyObject *key_str_map) {
 }
 /* }}} */
 
-static PyObject *_PylibMC_parse_memcached_value(char *value, size_t size,
-        uint32_t flags) {
+static PyObject *_PylibMC_parse_memcached_value(PylibMC_Client *self,
+        char *value, size_t size, uint32_t flags) {
     PyObject *retval = NULL;
-    PyObject *tmp = NULL;
-    uint32_t dtype = flags & PYLIBMC_FLAG_TYPES;
 
 #if USE_ZLIB
     PyObject *inflated = NULL;
@@ -551,65 +599,123 @@ static PyObject *_PylibMC_parse_memcached_value(char *value, size_t size,
     }
 #endif
 
-    switch (dtype) {
-        case PYLIBMC_FLAG_PICKLE:
-            retval = _PylibMC_Unpickle(value, size);
-            break;
-        case PYLIBMC_FLAG_INTEGER:
-        case PYLIBMC_FLAG_LONG:
-        case PYLIBMC_FLAG_BOOL:
-            /* PyInt_FromString doesn't take a length param and we're
-               not NULL-terminated, so we'll have to make an
-               intermediate Python string out of it */
-            tmp = PyBytes_FromStringAndSize(value, size);
-            if(tmp == NULL) {
-              goto cleanup;
-            }
-            retval = PyLong_FromString(PyBytes_AS_STRING(tmp), NULL, 10);
-            if(retval != NULL && dtype == PYLIBMC_FLAG_BOOL) {
-              Py_DECREF(tmp);
-              tmp = retval;
-              retval = PyBool_FromLong(PyLong_AS_LONG(tmp));
-            }
-            break;
-        case PYLIBMC_FLAG_NONE:
-            retval = PyBytes_FromStringAndSize(value, (Py_ssize_t)size);
-            break;
-        default:
-            PyErr_Format(PylibMCExc_Error,
-                    "unknown memcached key flags %u", flags);
+    if (self->native_deserialization) {
+        retval = _PylibMC_deserialize_native(self, NULL, value, size, flags);
+    } else {
+#if PY_MAJOR_VERSION >= 3
+        retval = PyObject_CallMethod((PyObject *)self, "deserialize", "y#I", value, size, (unsigned int) flags);
+#else
+        retval = PyObject_CallMethod((PyObject *)self, "deserialize", "s#I", value, size, (unsigned int) flags);
+#endif
     }
 
-cleanup:
-
 #if USE_ZLIB
     Py_XDECREF(inflated);
 #endif
 
-    Py_XDECREF(tmp);
 
     return retval;
 }
 
-static PyObject *_PylibMC_parse_memcached_result(memcached_result_st *res) {
-        return _PylibMC_parse_memcached_value((char *)memcached_result_value(res),
+/** Helper because PyLong_FromString requires a null-terminated string. */
+static PyObject *_PyLong_FromStringAndSize(char *value, size_t size, char **pend, int base) {
+    PyObject *retval;
+    char *tmp;
+    if ((tmp = malloc(size+1)) == NULL) {
+        return PyErr_NoMemory();
+    }
+    strncpy(tmp, value, size);
+    tmp[size] = '\0';
+    retval = PyLong_FromString(tmp, pend, base);
+    free(tmp);
+    return retval;
+}
+
+/** C implementation of deserialization.
+
+  This either takes a Python bytestring as `value`, or else `value` is NULL and
+  the value to be deserialized is a byte array `value_str` of length
+  `value_size`.
+  */
+static PyObject *_PylibMC_deserialize_native(PylibMC_Client *self, PyObject *value, char *value_str, size_t value_size, uint32_t flags) {
+    assert(value || value_str);
+    PyObject *retval = NULL;
+
+    uint32_t dtype = flags & PYLIBMC_FLAG_TYPES;
+
+    switch (dtype) {
+        case PYLIBMC_FLAG_PICKLE:
+            retval = value ? _PylibMC_Unpickle_Bytes(value) : _PylibMC_Unpickle(value_str, value_size);
+            break;
+        case PYLIBMC_FLAG_INTEGER:
+        case PYLIBMC_FLAG_LONG:
+        case PYLIBMC_FLAG_BOOL:
+            if (value) {
+                retval = PyLong_FromString(PyBytes_AS_STRING(value), NULL, 10);
+            } else {
+                retval = _PyLong_FromStringAndSize(value_str, value_size, NULL, 10);;
+            }
+            if (retval != NULL && dtype == PYLIBMC_FLAG_BOOL) {
+                PyObject *bool_retval = PyBool_FromLong(PyLong_AS_LONG(retval));
+                Py_DECREF(retval);
+                retval = bool_retval;
+            }
+            break;
+        case PYLIBMC_FLAG_NONE:
+            if (value) {
+                /* acquire an additional reference for parity */
+                Py_INCREF(value);
+                retval = value;
+            } else {
+                retval = PyBytes_FromStringAndSize(value_str, value_size);
+            }
+            break;
+        default:
+            PyErr_Format(PylibMCExc_Error,
+                    "unknown memcached key flags %u", dtype);
+    }
+
+    return retval;
+}
+
+static PyObject *PylibMC_Client_deserialize(PylibMC_Client *self, PyObject *args) {
+    PyObject *value;
+    unsigned int flags;
+    if (!PyArg_ParseTuple(args, "OI", &value, &flags)) {
+        return NULL;
+    }
+    return _PylibMC_deserialize_native(self, value, NULL, 0, flags);
+}
+
+static PyObject *_PylibMC_parse_memcached_result(PylibMC_Client *self, memcached_result_st *res) {
+        return _PylibMC_parse_memcached_value(
+                                              self,
+                                              (char *)memcached_result_value(res),
                                               memcached_result_length(res),
                                               memcached_result_flags(res));
 }
 
+/* Helper to call after _PylibMC_parse_memcached_value;
+   determines whether the deserialized value should be ignored
+   and treated as a miss.
+*/
+static int _PylibMC_cache_miss_simulated(PyObject *r) {
+    if (r == NULL && PyErr_Occurred() && PyErr_ExceptionMatches(PylibMCExc_CacheMiss)) {
+        PyErr_Clear();
+        return 1;
+    }
+    return 0;
+}
+
 static PyObject *PylibMC_Client_get(PylibMC_Client *self, PyObject *arg) {
     char *mc_val;
     size_t val_size;
     uint32_t flags;
     memcached_return error;
 
-    Py_INCREF(arg);
-
     if (!_key_normalized_obj(&arg)) {
-        Py_DECREF(arg);
         return NULL;
     } else if (!PySequence_Length(arg) ) {
-        Py_DECREF(arg);
         Py_RETURN_NONE;
     }
 
@@ -622,8 +728,13 @@ static PyObject *PylibMC_Client_get(PylibMC_Client *self, PyObject *arg) {
     Py_DECREF(arg);
 
     if (mc_val != NULL) {
-        PyObject *r = _PylibMC_parse_memcached_value(mc_val, val_size, flags);
+        PyObject *r = _PylibMC_parse_memcached_value(self, mc_val, val_size, flags);
         free(mc_val);
+        if (_PylibMC_cache_miss_simulated(r)) {
+            /* Since python-memcache returns None when the key doesn't exist,
+             * so shall we. */
+            Py_RETURN_NONE;
+        }
         return r;
     } else if (error == MEMCACHED_SUCCESS) {
         /* This happens for empty values, and so we fake an empty string. */
@@ -646,10 +757,7 @@ static PyObject *PylibMC_Client_gets(PylibMC_Client *self, PyObject *arg) {
     memcached_return rc;
     PyObject* ret = NULL;
 
-    Py_INCREF(arg);
-
     if (!_key_normalized_obj(&arg)) {
-        Py_DECREF(arg);
         return NULL;
     } else if (!PySequence_Length(arg)) {
         return Py_BuildValue("(OO)", Py_None, Py_None);
@@ -674,25 +782,37 @@ static PyObject *PylibMC_Client_gets(PylibMC_Client *self, PyObject *arg) {
 
     Py_END_ALLOW_THREADS;
 
+    int miss = 0;
+    int fail = 0;
     if (rc == MEMCACHED_SUCCESS && res != NULL) {
-        ret = Py_BuildValue("(NL)",
-                            _PylibMC_parse_memcached_result(res),
-                            memcached_result_cas(res));
+        PyObject *val = _PylibMC_parse_memcached_result(self, res);
+        if (_PylibMC_cache_miss_simulated(val)) {
+            miss = 1;
+        } else {
+            ret = Py_BuildValue("(NL)",
+                                val,
+                                memcached_result_cas(res));
+        }
 
         /* we have to fetch the last result from the mget cursor */
         if (NULL != memcached_fetch_result(self->mc, NULL, &rc)) {
             memcached_quit(self->mc);
             Py_DECREF(ret);
             ret = NULL;
+            fail = 1;
             PyErr_SetString(PyExc_RuntimeError, "fetch not done");
         }
     } else if (rc == MEMCACHED_END || rc == MEMCACHED_NOTFOUND) {
-        /* Key not found => (None, None) */
-        ret = Py_BuildValue("(OO)", Py_None, Py_None);
+        miss = 1;
     } else {
         ret = PylibMC_ErrFromMemcached(self, "memcached_gets", rc);
     }
 
+    if (miss && !fail) {
+        /* Key not found => (None, None) */
+        ret = Py_BuildValue("(OO)", Py_None, Py_None);
+    }
+
     if (res != NULL) {
         memcached_result_free(res);
     }
@@ -768,7 +888,7 @@ static PyObject *_PylibMC_RunSetCommandSingle(PylibMC_Client *self,
      */
     key = PyBytes_FromStringAndSize(key_raw, keylen);
 
-    success = _PylibMC_SerializeValue(key, NULL, value, time, &serialized);
+    success = _PylibMC_SerializeValue(self, key, NULL, value, time, &serialized);
 
     if (!success)
         goto cleanup;
@@ -837,7 +957,7 @@ static PyObject *_PylibMC_RunSetCommandMulti(PylibMC_Client *self,
 
     nkeys = (size_t)PyDict_Size(keys);
 
-    key_str_map = _PylibMC_map_str_keys(keys);
+    key_str_map = _PylibMC_map_str_keys(keys, NULL, NULL);
     if (key_str_map == NULL) {
         goto cleanup;
     }
@@ -852,7 +972,7 @@ static PyObject *_PylibMC_RunSetCommandMulti(PylibMC_Client *self,
     }
 
     for (i = 0, idx = 0; PyDict_Next(keys, &i, &curr_key, &curr_value); idx++) {
-        int success = _PylibMC_SerializeValue(curr_key, key_prefix,
+        int success = _PylibMC_SerializeValue(self, curr_key, key_prefix,
                                               curr_value, time,
                                               &serialized[idx]);
 
@@ -933,7 +1053,7 @@ static PyObject *_PylibMC_RunCasCommand(PylibMC_Client *self,
 
     /* TODO: because it's RunSetCommand that does the zlib
        compression, cas can't currently use compressed values. */
-    success = _PylibMC_SerializeValue(key, NULL, value, time, &mset);
+    success = _PylibMC_SerializeValue(self, key, NULL, value, time, &mset);
 
     if (!success || PyErr_Occurred() != NULL) {
         goto cleanup;
@@ -979,25 +1099,21 @@ static void _PylibMC_FreeMset(pylibmc_mset *mset) {
     mset->value_obj = NULL;
 }
 
-static int _PylibMC_SerializeValue(PyObject* key_obj,
+static int _PylibMC_SerializeValue(PylibMC_Client *self,
+                                   PyObject* key_obj,
                                    PyObject* key_prefix,
                                    PyObject* value_obj,
                                    time_t time,
                                    pylibmc_mset* serialized) {
-    PyObject* store_val = NULL;
 
     /* first zero the whole structure out */
     memset((void *)serialized, 0x0, sizeof(pylibmc_mset));
 
     serialized->time = time;
     serialized->success = false;
-    serialized->flags = PYLIBMC_FLAG_NONE;
-
-    /* Make an owned reference to key_obj */
-    Py_INCREF(key_obj);
+    serialized->value_obj = NULL;
 
     if (!_key_normalized_obj(&key_obj)) {
-        Py_DECREF(key_obj);
         return false;
     }
 
@@ -1012,10 +1128,7 @@ static int _PylibMC_SerializeValue(PyObject* key_obj,
 
     /* Check the key_prefix */
     if (key_prefix != NULL) {
-        Py_INCREF(key_prefix);
-
         if (!_key_normalized_obj(&key_prefix)) {
-            Py_DECREF(key_prefix);
             return false;
         }
 
@@ -1052,32 +1165,101 @@ static int _PylibMC_SerializeValue(PyObject* key_obj,
         serialized->prefixed_key_obj = prefixed_key_obj;
     }
 
-    /* Build store_val, a Python str/bytes object */
+    int success;
+    /* Build serialized->value_obj, a Python str/bytes object. */
+    if (self->native_serialization) {
+        success = _PylibMC_serialize_native(self, value_obj, &(serialized->value_obj), &(serialized->flags));
+    } else {
+        success = _PylibMC_serialize_user(self, value_obj, &(serialized->value_obj), &(serialized->flags));
+    }
+
+    if (!success) {
+        return false;
+    }
+
+    if (PyBytes_AsStringAndSize(serialized->value_obj, &serialized->value,
+                                &serialized->value_len) == -1) {
+        return false;
+    }
+
+    return true;
+}
+
+static int _PylibMC_serialize_user(PylibMC_Client *self, PyObject *value_obj, PyObject **dest, uint32_t *flags) {
+    PyObject *serval_and_flags = PyObject_CallMethod((PyObject *)self, "serialize", "(O)", value_obj);
+    if (serval_and_flags == NULL) {
+        return false;
+    }
+
+    if (PyTuple_Check(serval_and_flags)) {
+        PyObject *flags_obj = PyTuple_GetItem(serval_and_flags, 1);
+        if (flags_obj != NULL) {
+#if PY_MAJOR_VERSION >= 3
+            if (PyLong_Check(flags_obj)) {
+                *flags = (uint32_t) PyLong_AsLong(flags_obj);
+                *dest = PyTuple_GetItem(serval_and_flags, 0);
+            }
+#else
+            if (PyInt_Check(flags_obj)) {
+                *flags = (uint32_t) PyInt_AsLong(flags_obj);
+                *dest = PyTuple_GetItem(serval_and_flags, 0);
+            }
+#endif
+        }
+    }
+
+    if (*dest == NULL) {
+        /* PyErr_SetObject(PyExc_ValueError, serval_and_flags); */
+        PyErr_SetString(PyExc_ValueError, "serialize() must return (bytes, flags)");
+        Py_DECREF(serval_and_flags);
+        return false;
+    } else {
+        /* We're getting rid of serval_and_flags, which owns the only new
+           reference to value_obj. However, we can't deallocate value_obj
+           until we're done with value and value_len (after the set
+           operation). Therefore, take possession of a new reference to it
+           before cleaning up the tuple: */
+        Py_INCREF(*dest);
+        Py_DECREF(serval_and_flags);
+    }
+
+    return true;
+}
+
+/** C implementation of serialization.
+
+    This either returns true and stores the serialized bytestring in *dest
+    and the flags in *flags, or it returns false with an exception set.
+*/
+static int _PylibMC_serialize_native(PylibMC_Client *self, PyObject *value_obj, PyObject **dest, uint32_t *flags) {
+    PyObject *store_val = NULL;
+    uint32_t store_flags = PYLIBMC_FLAG_NONE;
 
     if (PyBytes_Check(value_obj)) {
+        store_flags = PYLIBMC_FLAG_NONE;
         /* Make store_val an owned reference */
         store_val = value_obj;
         Py_INCREF(store_val);
+    } else if (PyBool_Check(value_obj)) {
+        store_flags |= PYLIBMC_FLAG_BOOL;
+        /* bool cannot be subclassed; there are only two singleton values,
+           Py_True and Py_False */
+        const char *value_str = (value_obj == Py_True) ? "1" : "0";
+        store_val = PyBytes_FromString(value_str);
 #if PY_MAJOR_VERSION >= 3
-    } else if (PyBool_Check(value_obj)) {
-        serialized->flags |= PYLIBMC_FLAG_BOOL;
-        store_val = PyBytes_FromFormat("%ld", PyLong_AsLong(value_obj));
     } else if (PyLong_Check(value_obj)) {
-        serialized->flags |= PYLIBMC_FLAG_LONG;
-        store_val = PyBytes_FromFormat("%ld", PyLong_AsLong(value_obj));
-#else
-    } else if (PyBool_Check(value_obj)) {
-        serialized->flags |= PYLIBMC_FLAG_BOOL;
-        PyObject* tmp = PyNumber_Long(value_obj);
-        store_val = PyObject_Bytes(tmp);
+        store_flags |= PYLIBMC_FLAG_LONG;
+        PyObject *tmp = PyObject_Str(value_obj);
+        store_val = PyUnicode_AsEncodedString(tmp, "ascii", "strict");
         Py_DECREF(tmp);
+#else
     } else if (PyInt_Check(value_obj)) {
-        serialized->flags |= PYLIBMC_FLAG_INTEGER;
+        store_flags |= PYLIBMC_FLAG_INTEGER;
         PyObject* tmp = PyNumber_Int(value_obj);
         store_val = PyObject_Bytes(tmp);
         Py_DECREF(tmp);
     } else if (PyLong_Check(value_obj)) {
-        serialized->flags |= PYLIBMC_FLAG_LONG;
+        store_flags |= PYLIBMC_FLAG_LONG;
         PyObject* tmp = PyNumber_Long(value_obj);
         store_val = PyObject_Bytes(tmp);
         Py_DECREF(tmp);
@@ -1085,7 +1267,7 @@ static int _PylibMC_SerializeValue(PyObject* key_obj,
     } else if (value_obj != NULL) {
         /* we have no idea what it is, so we'll store it pickled */
         Py_INCREF(value_obj);
-        serialized->flags |= PYLIBMC_FLAG_PICKLE;
+        store_flags |= PYLIBMC_FLAG_PICKLE;
         store_val = _PylibMC_Pickle(value_obj);
         Py_DECREF(value_obj);
     }
@@ -1094,18 +1276,21 @@ static int _PylibMC_SerializeValue(PyObject* key_obj,
         return false;
     }
 
-    /* store_val is an owned reference; released when we're done with value and
-     * value_len (i.e. not here.) */
-    serialized->value_obj = store_val;
-
-    if (PyBytes_AsStringAndSize(store_val, &serialized->value,
-                                &serialized->value_len) == -1) {
-        return false;
-    }
-
+    *dest = store_val;
+    *flags = store_flags;
     return true;
 }
 
+static PyObject *PylibMC_Client_serialize(PylibMC_Client *self, PyObject *value_obj) {
+    PyObject *serialized_val;
+    uint32_t flags;
+    if (!_PylibMC_serialize_native(self, value_obj, &serialized_val, &flags)) {
+        return NULL;
+    }
+    /* we own a reference to serialized_val. "give" it to the tuple return value: */
+    return Py_BuildValue("(NI)", serialized_val, flags);
+}
+
 /* {{{ Set commands (set, replace, add, prepend, append) */
 static bool _PylibMC_RunSetCommand(PylibMC_Client* self,
                                    _PylibMC_SetCommand f, char *fname,
@@ -1562,17 +1747,17 @@ memcached_return pylibmc_memcached_fetch_multi(memcached_st *mc, pylibmc_mget_re
 
 static PyObject *PylibMC_Client_get_multi(
         PylibMC_Client *self, PyObject *args, PyObject *kwds) {
-    PyObject *key_seq, **key_objs, *retval = NULL;
+    PyObject *key_seq, **key_objs, **orig_key_objs, *retval = NULL;
     char **keys, *prefix = NULL;
     char *err_func = NULL;
     memcached_result_st *res, *results = NULL;
     Py_ssize_t prefix_len = 0;
     Py_ssize_t i;
-    PyObject *key_it, *ckey;
     PyObject *key_str_map = NULL;
     PyObject *temp_key_obj;
     size_t *key_lens;
-    size_t nkeys, nresults = 0;
+    Py_ssize_t nkeys = 0, orig_nkeys = 0;
+    size_t nresults = 0;
     memcached_return rc;
     pylibmc_mget_req req;
 
@@ -1582,46 +1767,50 @@ static PyObject *PylibMC_Client_get_multi(
             &key_seq, &prefix, &prefix_len))
         return NULL;
 
-    if ((nkeys = (size_t)PySequence_Length(key_seq)) == -1)
+    if ((orig_nkeys = PySequence_Length(key_seq)) == -1)
         return NULL;
 
     /* Populate keys and key_lens. */
-    keys = PyMem_New(char *, nkeys);
-    key_lens = PyMem_New(size_t, nkeys);
-    key_objs = PyMem_New(PyObject *, nkeys);
-    if (!keys || !key_lens || !key_objs) {
-        PyMem_Free(keys);
-        PyMem_Free(key_lens);
-        PyMem_Free(key_objs);
-        return PyErr_NoMemory();
+    keys = PyMem_New(char *, orig_nkeys);
+    key_lens = PyMem_New(size_t, (size_t) orig_nkeys);
+    key_objs = PyMem_New(PyObject *, (size_t) orig_nkeys);
+    orig_key_objs = PyMem_New(PyObject *, (size_t) orig_nkeys);
+    if (!keys || !key_lens || !key_objs || !orig_key_objs) {
+        PyErr_NoMemory();
+        goto memory_cleanup;
     }
 
     /* Clear potential previous exception, because we explicitly check for
      * exceptions as a loop predicate. */
     PyErr_Clear();
 
-    key_str_map = _PylibMC_map_str_keys(key_seq);
+    key_str_map = _PylibMC_map_str_keys(key_seq, orig_key_objs, &orig_nkeys);
+    if (key_str_map == NULL) {
+        goto memory_cleanup;
+    }
+
     /* Iterate through all keys and set lengths etc. */
-    key_it = PyObject_GetIter(key_seq);
-    i = 0;
-    while ((ckey = PyIter_Next(key_it)) != NULL) {
+    Py_ssize_t key_idx = 0;
+    for (i = 0; i < orig_nkeys; i++) {
+        PyObject *ckey = orig_key_objs[i];
         char *key;
         Py_ssize_t key_len;
+        size_t final_key_len;
         PyObject *rkey;
 
-        assert(i < nkeys);
-
         if (PyErr_Occurred() || !_key_normalized_obj(&ckey)) {
-            nkeys = i;
+            nkeys = key_idx;
             goto earlybird;
         }
 
+        /* normalization created an owned reference to ckey */
+
         PyBytes_AsStringAndSize(ckey, &key, &key_len);
 
-        key_lens[i] = (size_t)(key_len + prefix_len);
+        final_key_len = (size_t)(key_len + prefix_len);
 
         /* Skip empty keys */
-        if (!key_lens[i]) {
+        if (!final_key_len) {
             Py_DECREF(ckey);
             continue;
         }
@@ -1641,11 +1830,13 @@ static PyObject *PylibMC_Client_get_multi(
         }
         Py_DECREF(ckey);
 
-        keys[i] = PyBytes_AS_STRING(rkey);
-        key_objs[i++] = rkey;
+        keys[key_idx] = PyBytes_AS_STRING(rkey);
+        key_objs[key_idx] = rkey;
+        key_lens[key_idx] = final_key_len;
+        key_idx++;
+        /* we've added a total of 1 ref, to rkey */
     }
-    nkeys = i;
-    Py_XDECREF(key_it);
+    nkeys = key_idx;
 
     if (nkeys == 0) {
         retval = PyDict_New();
@@ -1664,7 +1855,7 @@ static PyObject *PylibMC_Client_get_multi(
     Py_BEGIN_ALLOW_THREADS;
 
     req.keys = keys;
-    req.nkeys = nkeys;
+    req.nkeys = (ssize_t) nkeys;
     req.key_lens = key_lens;
     req.results = &results;
     req.nresults = &nresults;
@@ -1695,15 +1886,22 @@ static PyObject *PylibMC_Client_get_multi(
         if (PyDict_Contains(key_str_map, key_obj)) {
             temp_key_obj = key_obj;
             key_obj = PyDict_GetItem(key_str_map, temp_key_obj);
+            /* PyDict_GetItem borrows a reference; own it for consistency */
+            Py_INCREF(key_obj);
             Py_DECREF(temp_key_obj);
         }
 
         /* Parse out value */
-        val = _PylibMC_parse_memcached_result(res);
-        if (val == NULL)
+        val = _PylibMC_parse_memcached_result(self, res);
+        if (_PylibMC_cache_miss_simulated(val)) {
+            Py_DECREF(key_obj);
+            continue;
+        }
+        else if (val == NULL)
             goto loopcleanup;
 
         rc = PyDict_SetItem(retval, key_obj, val);
+        /* clean up our local owned references (now that rc has its own) */
         Py_DECREF(key_obj);
         Py_DECREF(val);
 
@@ -1719,14 +1917,18 @@ loopcleanup:
     }
 
 earlybird:
-    PyMem_Free(key_lens);
-    PyMem_Free(keys);
-
+    for (i = 0; i < orig_nkeys; i++)
+        Py_DECREF(orig_key_objs[i]);
     for (i = 0; i < nkeys; i++)
         Py_DECREF(key_objs[i]);
-    PyMem_Free(key_objs);
     _PylibMC_cleanup_str_key_mapping(key_str_map);
 
+memory_cleanup:
+    PyMem_Free(key_lens);
+    PyMem_Free(keys);
+    PyMem_Free(key_objs);
+    PyMem_Free(orig_key_objs);
+
     if (results != NULL) {
         for (i = 0; i < nresults && results != NULL; i++) {
             memcached_result_free(results + i);
@@ -1752,6 +1954,7 @@ static PyObject *_PylibMC_DoMulti(PyObject *values, PyObject *func,
     PyObject *retval = PyList_New(0);
     PyObject *iter = NULL;
     PyObject *item = NULL;
+    PyObject *orig_item = NULL;
     int is_mapping = PyDict_Check(values);
 
     if (retval == NULL)
@@ -1760,7 +1963,7 @@ static PyObject *_PylibMC_DoMulti(PyObject *values, PyObject *func,
     if ((iter = PyObject_GetIter(values)) == NULL)
         goto cleanup;
 
-    while ((item = PyIter_Next(iter)) != NULL) {
+    while ((orig_item = item = PyIter_Next(iter)) != NULL) {
         PyObject *args_f = NULL;
         PyObject *args = NULL;
         PyObject *key = NULL;
@@ -1775,18 +1978,19 @@ static PyObject *_PylibMC_DoMulti(PyObject *values, PyObject *func,
         if (!_key_normalized_obj(&item))
             goto loopcleanup;
         if (prefix == NULL || prefix == Py_None) {
-            /* We now have two owned references to item. */
             key = item;
             Py_INCREF(key);
         } else {
-            key = PySequence_Concat(prefix, item);
+            PyObject *concat_key = key = PySequence_Concat(prefix, item);
+            /**
+             * Another call to _key_normalized_obj is still a good idea since
+             * we might have created a key that's too long
+             */
+            if (key == NULL || !_key_normalized_obj(&key))
+                goto loopcleanup;
+            /* Normalization created a new ref to key; clean up the temporary ref. */
+            Py_DECREF(concat_key);
         }
-        /**
-         * Another call to _key_normalized_obj is still a good idea since
-         * we might have created a key that's too long
-         */
-        if (key == NULL || !_key_normalized_obj(&key))
-            goto loopcleanup;
 
         /* Calculate args. */
         if (is_mapping) {
@@ -1828,6 +2032,7 @@ static PyObject *_PylibMC_DoMulti(PyObject *values, PyObject *func,
         Py_DECREF(args);
         Py_DECREF(key);
         Py_DECREF(item);
+        Py_DECREF(orig_item);
         continue;
 loopcleanup:
         Py_XDECREF(args_f);
@@ -2202,6 +2407,8 @@ static PyObject *PylibMC_Client_clone(PylibMC_Client *self) {
     Py_BEGIN_ALLOW_THREADS;
     clone->mc = memcached_clone(NULL, self->mc);
     Py_END_ALLOW_THREADS;
+    clone->native_serialization = self->native_serialization;
+    clone->native_deserialization = self->native_deserialization;
     return (PyObject *)clone;
 }
 /* }}} */
@@ -2283,64 +2490,84 @@ static PyObject *_PylibMC_GetPickles(const char *attname) {
 }
 
 static PyObject *_PylibMC_Unpickle(const char *buff, size_t size) {
-    PyObject *pickle_load;
-    PyObject *retval = NULL;
-
-    retval = NULL;
-    pickle_load = _PylibMC_GetPickles("loads");
-    if (pickle_load != NULL) {
 #if PY_MAJOR_VERSION >= 3
-        retval = PyObject_CallFunction(pickle_load, "y#", buff, size);
+        return PyObject_CallFunction(_PylibMC_pickle_loads, "y#", buff, size);
 #else
-        retval = PyObject_CallFunction(pickle_load, "s#", buff, size);
+        return PyObject_CallFunction(_PylibMC_pickle_loads, "s#", buff, size);
 #endif
-        Py_DECREF(pickle_load);
-    }
+}
 
-    return retval;
+static PyObject *_PylibMC_Unpickle_Bytes(PyObject *val) {
+    return PyObject_CallFunctionObjArgs(_PylibMC_pickle_loads, val, NULL);
 }
 
 static PyObject *_PylibMC_Pickle(PyObject *val) {
-    PyObject *pickle_dump;
-    PyObject *retval = NULL;
-
-    pickle_dump = _PylibMC_GetPickles("dumps");
-    if (pickle_dump != NULL) {
-        retval = PyObject_CallFunction(pickle_dump, "Oi", val, -1);
-        Py_DECREF(pickle_dump);
-    }
-
-    return retval;
+    return PyObject_CallFunctionObjArgs(_PylibMC_pickle_dumps, val, NULL);
 }
 /* }}} */
 
+/**
+ * Normalize a key. Takes in a PyObject **.
+ *
+ * Returns a status code; on success (code != 0), *key will contain a new
+ * reference to a normalized key (possibly the original object itself).
+ * On error (code == 0), no references are created.
+ */
 static int _key_normalized_obj(PyObject **key) {
     int rc;
     char *key_str;
     Py_ssize_t key_sz;
+    PyObject *orig_key = *key;
+    PyObject *retval = orig_key;
+    PyObject *encoded_key = NULL;
 
     if (*key == NULL) {
         PyErr_SetString(PyExc_ValueError, "key must be given");
         return 0;
     }
 
-    if (PyUnicode_Check(*key)) {
-        *key = PyUnicode_AsUTF8String(*key);
-        if (*key == NULL)
-            return 0;
+    /* The key object may be borrowed; hold a ref to it for now. */
+    Py_INCREF(orig_key);
+
+    if (PyUnicode_Check(retval)) {
+        encoded_key = PyUnicode_AsUTF8String(retval);
+        retval = encoded_key;
+        if (retval == NULL) {
+            rc = 0;
+            goto END;
+        }
     }
 
-    if (!PyBytes_Check(*key)) {
+    if (!PyBytes_Check(retval)) {
         PyErr_SetString(PyExc_TypeError, "key must be bytes");
-        return 0;
+        rc = 0;
+        retval = NULL;
+        goto END;
     }
 
-    key_str = PyBytes_AS_STRING(*key);
-    key_sz = PyBytes_GET_SIZE(*key);
+    key_str = PyBytes_AS_STRING(retval);
+    key_sz = PyBytes_GET_SIZE(retval);
     rc = _key_normalized_str(&key_str, &key_sz);
     if (rc == 2) {
-        *key = PyBytes_FromStringAndSize(key_str, key_sz);
-        rc = 1;
+        retval = PyBytes_FromStringAndSize(key_str, key_sz);
+        if (retval != NULL) {
+            rc = 1;
+        } else {
+            rc = 0;
+        }
+    }
+
+END:
+    if (retval != orig_key) {
+        /* If retval is new, release the temporary reference on the original. */
+        Py_DECREF(orig_key);
+    }
+    if (retval != encoded_key) {
+        /* Same deal for the UTF-8 encoded key, if we made one. */
+        Py_XDECREF(encoded_key);
+    }
+    if (retval != NULL) {
+        *key = retval;
     }
     return rc;
 }
@@ -2432,9 +2659,14 @@ static void _make_excs(PyObject *module) {
     PylibMCExc_Error = PyErr_NewException(
             "pylibmc.Error", NULL, NULL);
 
+    PylibMCExc_CacheMiss = PyErr_NewException(
+            "_pylibmc.CacheMiss", PylibMCExc_Error, NULL);
+
     exc_objs = PyList_New(0);
     PyList_Append(exc_objs,
                   Py_BuildValue("sO", "Error", (PyObject *)PylibMCExc_Error));
+    PyList_Append(exc_objs,
+                  Py_BuildValue("sO", "CacheMiss", (PyObject *)PylibMCExc_CacheMiss));
 
     for (err = PylibMCExc_mc_errs; err->name != NULL; err++) {
         char excnam[64];
@@ -2449,6 +2681,9 @@ static void _make_excs(PyObject *module) {
     PyModule_AddObject(module, "Error",
                        (PyObject *)PylibMCExc_Error);
 
+    PyModule_AddObject(module, "CacheMiss",
+                       (PyObject *)PylibMCExc_CacheMiss);
+
     /* Backwards compatible name for <= pylibmc 1.2.3
      *
      * Need to increase the refcount since we're adding another
@@ -2536,6 +2771,14 @@ by using comma-separation. Good luck with that.\n", PylibMC_functions);
 
     _make_excs(module);
 
+    if (!(_PylibMC_pickle_loads = _PylibMC_GetPickles("loads"))) {
+        return MOD_ERROR_VAL;
+    }
+
+    if (!(_PylibMC_pickle_dumps = _PylibMC_GetPickles("dumps"))) {
+        return MOD_ERROR_VAL;
+    }
+
     PyModule_AddStringConstant(module, "__version__", PYLIBMC_VERSION);
     PyModule_ADD_REF(module, "client", (PyObject *)&PylibMC_ClientType);
     PyModule_AddStringConstant(module,
