@@ -131,6 +131,7 @@ typedef struct {
 
 /* {{{ Exceptions */
 static PyObject *PylibMCExc_Error;
+static PyObject *PylibMCExc_CacheMiss;
 
 /* Mapping of memcached_return value -> Python exception object. */
 typedef struct {
@@ -192,7 +193,9 @@ typedef struct {
 static PylibMC_Behavior PylibMC_behaviors[] = {
     { MEMCACHED_BEHAVIOR_NO_BLOCK, "no_block" },
     { MEMCACHED_BEHAVIOR_TCP_NODELAY, "tcp_nodelay" },
+#if LIBMEMCACHED_VERSION_HEX >= 0x00044000
     { MEMCACHED_BEHAVIOR_TCP_KEEPALIVE, "tcp_keepalive" },
+#endif
     { MEMCACHED_BEHAVIOR_HASH, "hash" },
     { MEMCACHED_BEHAVIOR_KETAMA_HASH, "ketama_hash" },
     { MEMCACHED_BEHAVIOR_KETAMA, "ketama" },
@@ -253,16 +256,16 @@ static PylibMC_Behavior PylibMC_hashers[] = {
 static PylibMC_Behavior PylibMC_distributions[] = {
     { MEMCACHED_DISTRIBUTION_MODULA, "modula" },
     { MEMCACHED_DISTRIBUTION_CONSISTENT, "consistent" },
+    { MEMCACHED_DISTRIBUTION_RANDOM, "random" },
+    { MEMCACHED_DISTRIBUTION_CONSISTENT_KETAMA, "consistent_ketama" },
+#if LIBMEMCACHED_VERSION_HEX >= 0x00044000
+    { MEMCACHED_DISTRIBUTION_CONSISTENT_KETAMA_SPY, "consistent_ketama_spy" },
+#endif
 #if LIBMEMCACHED_VERSION_HEX >= 0x01000000
     { MEMCACHED_DISTRIBUTION_CONSISTENT_WEIGHTED, "consistent_weighted" },
-#endif
-    { MEMCACHED_DISTRIBUTION_CONSISTENT_KETAMA, "consistent_ketama" },
-    { MEMCACHED_DISTRIBUTION_CONSISTENT_KETAMA_SPY, "consistent_ketama_spy" },
-    { MEMCACHED_DISTRIBUTION_RANDOM, "random" },
-#if LIBMEMCACHED_VERSION_HEX >= 0x01000000
     { MEMCACHED_DISTRIBUTION_VIRTUAL_BUCKET, "virtual_bucket" },
-#endif
     { MEMCACHED_DISTRIBUTION_CONSISTENT_MAX, "consistent_max" },
+#endif
     { 0, NULL }
 };
 /* }}} */
@@ -272,6 +275,8 @@ typedef struct {
     PyObject_HEAD
     memcached_st *mc;
     uint8_t sasl_set;
+    uint8_t native_serialization;
+    uint8_t native_deserialization;
 } PylibMC_Client;
 
 /* {{{ Prototypes */
@@ -279,6 +284,8 @@ static PylibMC_Client *PylibMC_ClientType_new(PyTypeObject *, PyObject *,
         PyObject *);
 static void PylibMC_ClientType_dealloc(PylibMC_Client *);
 static int PylibMC_Client_init(PylibMC_Client *, PyObject *, PyObject *);
+static PyObject *PylibMC_Client_deserialize(PylibMC_Client *, PyObject *arg);
+static PyObject *PylibMC_Client_serialize(PylibMC_Client *, PyObject *val);
 static PyObject *PylibMC_Client_get(PylibMC_Client *, PyObject *arg);
 static PyObject *PylibMC_Client_gets(PylibMC_Client *, PyObject *arg);
 static PyObject *PylibMC_Client_set(PylibMC_Client *, PyObject *, PyObject *);
@@ -308,10 +315,15 @@ static PyObject *PylibMC_ErrFromMemcachedWithKey(PylibMC_Client *, const char *,
 static PyObject *PylibMC_ErrFromMemcached(PylibMC_Client *, const char *,
         memcached_return);
 static PyObject *_PylibMC_Unpickle(const char *, size_t);
+static PyObject *_PylibMC_Unpickle_Bytes(PyObject *);
 static PyObject *_PylibMC_Pickle(PyObject *);
 static int _key_normalized_obj(PyObject **);
 static int _key_normalized_str(char **, Py_ssize_t *);
-static int _PylibMC_SerializeValue(PyObject *key_obj,
+static int _PylibMC_serialize_user(PylibMC_Client *, PyObject *, PyObject **, uint32_t *);
+static int _PylibMC_serialize_native(PylibMC_Client *, PyObject *, PyObject **, uint32_t *);
+static PyObject *_PylibMC_deserialize_native(PylibMC_Client *, PyObject *, char *, size_t, uint32_t);
+static int _PylibMC_SerializeValue(PylibMC_Client *self,
+                                   PyObject *key_obj,
                                    PyObject *key_prefix,
                                    PyObject *value_obj,
                                    time_t time,
@@ -338,6 +350,12 @@ static bool _PylibMC_IncrDecr(PylibMC_Client *, pylibmc_incr *, size_t);
 
 /* {{{ Type's method table */
 static PyMethodDef PylibMC_ClientType_methods[] = {
+    {"serialize", (PyCFunction)PylibMC_Client_serialize, METH_O,
+        "Serialize an object to a byte string and flag field, to be stored "
+         "in memcached."},
+    {"deserialize", (PyCFunction)PylibMC_Client_deserialize, METH_VARARGS,
+        "Deserialize a bytestring and flag field retrieved from memcached. "
+        "Raise pylibmc.CacheMiss to simulate a cache miss."},
     {"get", (PyCFunction)PylibMC_Client_get, METH_O,
         "Retrieve a key from a memcached."},
     {"gets", (PyCFunction)PylibMC_Client_gets, METH_O,
